"""Module containing the Data Store Product class."""

from __future__ import annotations

import re
from contextlib import contextmanager
from datetime import datetime
from functools import total_ordering
from typing import TYPE_CHECKING, Tuple, Dict

from eumdac.request import get
from eumdac.logging import logger

if TYPE_CHECKING:  # pragma: no cover
    import sys
    from typing import IO, Any, Optional

    if sys.version_info < (3, 9):
        from typing import Generator, Mapping, MutableMapping, Pattern, Sequence
    else:
        from collections.abc import Mapping, MutableMapping, Generator, Sequence
        from re import Pattern
    from eumdac.datastore import DataStore
    from eumdac.collection import Collection

from eumdac.errors import EumdacError, eumdac_raise_for_status
import eumdac.common


@total_ordering
class Product:
    """Product of a Collection in the Data Store

    Attributes:
    -----------
    - `datastore`:
        Reference to the Data Store
    - `download_url`:
        URL to the download endpoint of the Data Store

    Arguments:
    ----------
    - `collection_id`:
        Data Store ID of the collection
    - `product_id`:
        Data Store ID of the product
    - `datastore`:
        Reference to the Data Store
    """

    _id: str
    datastore: DataStore
    collection: Collection
    _browse_properties: Optional[Mapping[str, Any]] = None
    _download_properties: Optional[Mapping[str, Any]] = None
    _geometry: Optional[Mapping[str, Any]] = None
    _entries: Optional[Sequence[str]] = None
    _extract_filename: Pattern[str] = re.compile(r'filename="(.*?)"')
    _extract_sensing_time: Pattern[str] = re.compile(
        r"(?P<start>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?)?"
        r"\d*Z?[\\/]+"
        r"(?P<end>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?)?"
    )
    _extract_ingestion_time: Pattern[str] = re.compile(
        r"(?P<updated>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?)?"
    )

    def __init__(self, collection_id: str, product_id: str, datastore: DataStore) -> None:
        self._id = product_id
        self.datastore = datastore
        self.collection = self.datastore.get_collection(collection_id)

    def __str__(self) -> str:
        return str(self._id)

    def __repr__(self) -> str:
        return f"{self.__class__}({self.collection._id}, {self._id})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (
            self.collection._id,
            self._id,
        ) == (
            other.collection._id,
            other._id,
        )

    def __lt__(self, other: Product) -> bool:
        return (self.collection._id, self._id) < (other.collection._id, other._id)

    def __hash__(self) -> int:
        return hash(repr(self))

    def _parse_browse_properties(
        self, properties: dict[str, str | Any], geometry: Optional[dict[str, Any]]
    ) -> None:
        # Remove collection ID
        properties.pop("parentIdentifier")
        # Bring MD5 to base properties
        if "md5" in properties["extraInformation"]:
            properties.update({"md5": properties["extraInformation"]["md5"]})  # type: ignore
        else:
            properties.update({"md5": None})
        self._browse_properties = properties
        # Handle geometry
        if not geometry or len(geometry) == 0:
            self._geometry = None
        else:
            self._geometry = geometry
        # Handle entries
        self._entries = [link["title"] for link in properties["links"]["sip-entries"]]  # type: ignore

    def _ensure_properties(self) -> None:
        if self._browse_properties is not None:
            return
        url = self.datastore.urls.get(
            "datastore",
            "browse product",
            vars={"collection_id": self.collection._id, "product_id": self._id},
        )
        auth = self.datastore.token.auth
        response = get(
            url,
            params={"format": "json"},
            auth=auth,
            headers=eumdac.common.headers,
        )
        eumdac_raise_for_status(
            f"Product {self._id} not found in {self.collection._id}",
            response,
            ProductError,
        )

        # DSSHD-658: Better reporting of parsing errors
        try:
            response_json = response.json()
        except Exception as decode_error:
            logger.debug(f"Failed to parse JSON response: {response.text}")
            raise EumdacError(
                "Could not parse response received from server",
                {"response": response.text, "status_code": response.status_code},
            ) from decode_error

        properties = response_json["properties"]
        geometry: Optional[dict[str, Any]] = None
        if "geometry" in response_json:
            geometry = response_json["geometry"]

        self._parse_browse_properties(properties, geometry)

    def _ensure_download_properties(self) -> None:
        if self._download_properties is not None:
            return
        url = self.datastore.urls.get(
            "datastore",
            "download product metadata",
            vars={"collection_id": self.collection._id, "product_id": self._id},
        )
        auth = self.datastore.token.auth
        response = get(
            url,
            params={"format": "json"},
            auth=auth,
            headers=eumdac.common.headers,
        )
        eumdac_raise_for_status(
            f"Product {self._id} not found in {self.collection._id}",
            response,
            ProductError,
        )
        response_data = response.json()
        download_properties = response_data["properties"]
        self._download_properties = download_properties

    @property
    def sensing_start(self) -> datetime:
        """Sensing start date"""
        self._ensure_properties()
        try:
            sensing_time = self._extract_sensing_time.search(  # type: ignore[union-attr]
                self._browse_properties["date"]  # type: ignore[index]
            ).groupdict()
            sensing_start = sensing_time["start"]
            if len(sensing_start) > 19:
                sensing_start = sensing_start.ljust(26, "0")
            return datetime.fromisoformat(sensing_start)
        except Exception as e:
            raise ProductError(f"Couldn't prodcue sensing start time from received value: {self._browse_properties['date']}") from e  # type: ignore[index]

    @property
    def sensing_end(self) -> datetime:
        """Sensing end date"""
        self._ensure_properties()
        try:
            sensing_time = self._extract_sensing_time.search(  # type: ignore[union-attr]
                self._browse_properties["date"]  # type: ignore[index]
            ).groupdict()
            sensing_end = sensing_time["end"]
            if len(sensing_end) > 19:
                sensing_end = sensing_end.ljust(26, "0")
            return datetime.fromisoformat(sensing_end)
        except Exception as e:
            raise ProductError(f"Couldn't produce sensing end time from received value: {self._browse_properties['date']}") from e  # type: ignore[index]

    @property
    def satellite(self) -> str:
        """Platform or Mission related to the product"""
        self._ensure_properties()
        satellites = [
            acquisition["platform"]["platformShortName"]
            for acquisition in self._browse_properties["acquisitionInformation"]  # type: ignore[index]
        ]
        return ", ".join(satellites)

    @property
    def instrument(self) -> str:
        """Instrument related to the product"""
        self._ensure_properties()
        instruments = [
            acquisition["instrument"]["instrumentShortName"]
            for acquisition in self._browse_properties["acquisitionInformation"]  # type: ignore[index]
        ]
        return ", ".join(instruments)

    @property
    def size(self) -> int:
        """Size of the product"""
        self._ensure_properties()
        return self._browse_properties["productInformation"]["size"]  # type: ignore[index]

    @property
    def acronym(self) -> str:
        """Acronym or Product Type of the product"""
        self._ensure_properties()
        return self._browse_properties["productInformation"]["productType"]  # type: ignore[index]

    @property
    def product_type(self) -> str:
        """Product Type or Acronym of the product"""
        return self.acronym

    @property
    def timeliness(self) -> Optional[int]:
        """Timeliness of the product"""
        self._ensure_properties()
        if "timeliness" in self._browse_properties["productInformation"]:  # type: ignore[index]
            return self._browse_properties["productInformation"]["timeliness"]  # type: ignore[index]
        return None

    @property
    def md5(self) -> Optional[str]:
        """MD5 checksum of the product SIP"""
        self._ensure_properties()
        return self._browse_properties["md5"]  # type: ignore[index]

    @property
    def processingTime(self) -> Optional[str]:
        """Processing time"""
        self._ensure_download_properties()
        if "processingDate" in self._download_properties["productInformation"]["processingInformation"]:  # type: ignore[index]
            return self._download_properties["productInformation"]["processingInformation"]["processingDate"]  # type: ignore[index]
        return None

    @property
    def processorVersion(self) -> Optional[str]:  # type: ignore[valid-type]
        """Processor version"""
        self._ensure_download_properties()
        if "processorVersion" in self._download_properties["productInformation"]["processingInformation"]:  # type: ignore[index]
            return self._download_properties["productInformation"]["processingInformation"]["processorVersion"]  # type: ignore[index]
        return None

    @property
    def format(self) -> Optional[str]:
        """Format"""
        self._ensure_download_properties()
        if "format" in self._download_properties["productInformation"]["processingInformation"]:  # type: ignore[index]
            return self._download_properties["productInformation"]["processingInformation"]["format"]  # type: ignore[index]
        return None

    @property
    def qualityStatus(self) -> Optional[str]:
        """Quality status"""
        self._ensure_download_properties()
        if "qualityStatus" in self._download_properties["productInformation"]["qualityInformation"]:  # type: ignore[index]
            return self._download_properties["productInformation"]["qualityInformation"]["qualityStatus"]  # type: ignore[index]
        return None

    @property
    def ingested(self) -> Optional[datetime]:
        """Date of publication"""
        self._ensure_properties()
        ingestion_time = self._extract_ingestion_time.search(  # type: ignore[union-attr]
            self._browse_properties["updated"]  # type: ignore[index]
        ).groupdict()
        ingestion_time = ingestion_time["updated"]
        if len(ingestion_time) > 19:
            ingestion_time = ingestion_time.ljust(26, "0")
        return datetime.fromisoformat(ingestion_time)

    @property
    def orbit_type(self) -> str:
        """Orbit Type (GEO/LEO)"""
        self._ensure_download_properties()
        return self._download_properties["acquisitionInformation"][0]["platform"]["orbitType"]  # type: ignore[index]

    @property
    def orbit_is_LEO(self) -> bool:
        """Is Orbit Type LEO?"""
        self._ensure_download_properties()
        return self.orbit_type == "LEO"  # type: ignore[index]

    @property
    def orbit_number(self) -> Optional[int]:
        """Orbit Number of LEO product"""
        self._ensure_properties()
        if "orbitNumber" in self._browse_properties["acquisitionInformation"][0]["acquisitionParameters"]:  # type: ignore[index]
            return self._browse_properties["acquisitionInformation"][0]["acquisitionParameters"]["orbitNumber"]  # type: ignore[index]
        return None

    @property
    def orbit_direction(self) -> Optional[str]:
        """Orbit Direction of LEO product"""
        self._ensure_properties()
        if "orbitDirection" in self._browse_properties["acquisitionInformation"][0]["acquisitionParameters"]:  # type: ignore[index]
            return self._browse_properties["acquisitionInformation"][0]["acquisitionParameters"][  # type: ignore[index]
                "orbitDirection"
            ]
        return None

    @property
    def relative_orbit(self) -> Optional[int]:
        """Relative Orbit Number of LEO product"""
        self._ensure_properties()
        if (
            "relOrbitStart"
            in self._browse_properties["acquisitionInformation"][0]["acquisitionParameters"]  # type: ignore[index]
        ):
            return self._browse_properties["acquisitionInformation"][0]["acquisitionParameters"][  # type: ignore[index]
                "relOrbitStart"
            ]
        return None

    @property
    def cycle_number(self) -> Optional[int]:
        """Cycle Number of LEO product"""
        self._ensure_properties()
        if (
            "cycleNumber"
            in self._browse_properties["acquisitionInformation"][0]["acquisitionParameters"]  # type: ignore[index]
        ):
            return self._browse_properties["acquisitionInformation"][0]["acquisitionParameters"][  # type: ignore[index]
                "cycleNumber"
            ]
        return None

    @property
    def is_mtg(self) -> bool:
        """Is MTG collection?"""
        self._ensure_download_properties()
        return (
            "mtgCoverage"
            in self._download_properties["acquisitionInformation"][0]["acquisitionParameters"]  # type: ignore[index]
        )

    @property
    def repeat_cycle(self) -> Optional[str]:
        """Repeat cycle number (only for MTG products)"""
        self._ensure_properties()
        if (
            self.is_mtg
            and "repeatCycleIdentifier"
            in self._download_properties["acquisitionInformation"][0]["acquisitionParameters"]["mtgCoverage"]  # type: ignore[index]
        ):
            return self._download_properties["acquisitionInformation"][0]["acquisitionParameters"]["mtgCoverage"][  # type: ignore[index]
                "repeatCycleIdentifier"
            ]
        return None

    @property
    def region_coverage(self) -> Optional[str]:
        """Region Coverage (only for MTG products)"""
        self._ensure_properties()
        if (
            self.is_mtg
            and "majorRegionCoverage"
            in self._download_properties["acquisitionInformation"][0]["acquisitionParameters"]["mtgCoverage"]  # type: ignore[index]
        ):
            return self._download_properties["acquisitionInformation"][0]["acquisitionParameters"]["mtgCoverage"][  # type: ignore[index]
                "majorRegionCoverage"
            ]
        return None

    @property
    def subregion_identifier(self) -> Optional[str]:
        """Sub-Region (only for MTG products)"""
        self._ensure_properties()
        if (
            self.is_mtg
            and "subRegionIdentifier"
            in self._download_properties["acquisitionInformation"][0]["acquisitionParameters"]["mtgCoverage"]  # type: ignore[index]
        ):
            return self._download_properties["acquisitionInformation"][0]["acquisitionParameters"]["mtgCoverage"][  # type: ignore[index]
                "subRegionIdentifier"
            ]
        return None

    @property
    def metadata(self) -> MutableMapping[str, Any]:
        """Product metadata"""
        self._ensure_properties()
        self._ensure_download_properties()
        return {
            "geometry": self._geometry.copy() if self._geometry else None,  # type: ignore
            "properties": self._browse_properties.copy(),  # type: ignore[union-attr]
            "download_properties": self._download_properties.copy(),  # type: ignore[union-attr]
        }

    @property
    def entries(self) -> Sequence[str]:
        """Files inside the product"""
        self._ensure_properties()
        if not self._entries:
            return []
        else:
            return tuple(self._entries)

    @property
    def url(self) -> str:
        """URL of the product"""
        access_token = None
        if hasattr(self.datastore.token, "access_token"):
            access_token = str(self.datastore.token.access_token)

        url = self.datastore.urls.get(
            "datastore",
            "download product",
            vars={"collection_id": self.collection._id, "product_id": self._id},
        )
        download_url = str(url)
        if access_token is not None:
            download_url += "?access_token=" + access_token

        return download_url

    @contextmanager
    def open(
        self,
        entry: Optional[str] = None,
        chunk: Optional[Tuple[int, int]] = None,
        custom_headers: Optional[Dict[str, str]] = None,
    ) -> Generator[IO[bytes], None, None]:
        """Opens a stream to download the product content.

        Note:
            A Data Store product refers to a zip archive containing the data.

        Arguments:
            - `entry` (optional):
                specific file inside the product
            - `chunk` (optional):
                Byte range to download as (start, end). If None, downloads the full content.
            - `custom_headers` (optional):
                Additional HTTP headers to include in the request.

        Yields:
            Generator[IO[bytes]]: Chunks of the downloaded content.
        """
        url = self.datastore.urls.get(
            "datastore",
            "download product",
            vars={"collection_id": self.collection._id, "product_id": self._id},
        )
        auth = self.datastore.token.auth
        params = None
        if entry is not None:
            url += "/entry"
            params = {"name": entry}

        headers = eumdac.common.headers.copy()
        if chunk:
            headers = {**headers, "Range": f"bytes={chunk[0]}-{chunk[1]-1}"}
        if custom_headers:
            headers.update(custom_headers)

        with get(
            url,
            auth=auth,
            params=params,
            stream=True,
            headers=headers,
        ) as response:
            eumdac_raise_for_status(
                f"Could not download Product {self._id} of Collection {self.collection._id}",
                response,
                ProductError,
            )
            match = self._extract_filename.search(response.headers["Content-Disposition"])
            filename = match.group(1)  # type: ignore[union-attr]
            response.raw.name = filename
            response.raw.decode_content = True
            yield response.raw


class ProductError(EumdacError):
    """Errors related to products"""



import pytest
from datetime import timedelta
from eumdac.customisation import CustomisationError
from eumdac.tailor_app import (
    WaitForDoneCustomisationFutureFunc,
    InvalidStateTransitionError,
    DownloadRunFutureFunc,
    StateQueryFutureFunc,
    FatalEumdacError,
    ExceedingNumberOfCustomisationsEumdacError,
    TailorApp,
)
from eumdac.errors import EumdacError
from eumdac.job_id import JobIdentifier

from tests.base import FakeProduct, FakeCollection, FakeCustomisation, FakeTailor, FakeStore
from tests.test_order import tailor_order, products


class MockOrder:
    def __init__(self):
        self._no_warning_logs = True
        pass

    def initialize(self, chain, products, output_dir, no_warning_logs):
        pass

    def __str__(self):
        return "test"

    def deserialize(self):
        return {"max_workers": 3}

    def serialize(self, d):
        pass

    def update(self, customisation_id, product_id, status=None, download_states=None):
        pass

    def resolve_product_num(self, p_id):
        return -1, -1


valid_transitions = [
    ["RUNNING", "DONE"],
    ["QUEUED", "RUNNING", "DONE"],
    ["QUEUED", "QUEUED", "QUEUED", "RUNNING", "DONE"],
    ["QUEUED", "QUEUED", "QUEUED", "RUNNING", "RUNNING", "RUNNING", "DONE", "DONE"],
]


@pytest.mark.parametrize("transitions", valid_transitions)
def test_valid_transitions(transitions):
    def add_func(product):
        return FakeCustomisation(states_to_return=transitions)

    WaitForDoneCustomisationFutureFunc(
        customization_add_func=add_func,
        order=MockOrder(),
        polling_interval=0,
        job_identificator=JobIdentifier(total_jobs=1),
    )(FakeProduct())


invalid_transitions = [
    ["RUNNING", "QUEUED"],
]


@pytest.mark.parametrize("transitions", invalid_transitions)
def test_invalid_transitions(transitions):
    with pytest.raises(InvalidStateTransitionError):

        def add_func(product):
            return FakeCustomisation(states_to_return=transitions)

        WaitForDoneCustomisationFutureFunc(
            customization_add_func=add_func,
            order=MockOrder(),
            polling_interval=0,
            job_identificator=JobIdentifier(total_jobs=1),
        )(FakeProduct())


from itertools import cycle

timeout_transitions = [
    cycle(["QUEUED"]),
    cycle(["RUNNING"]),
]


@pytest.mark.parametrize("transitions", timeout_transitions)
def test_timeout_raised(transitions):
    with pytest.raises(FatalEumdacError):

        def mock_add_func(product):
            return FakeCustomisation(transitions)

        WaitForDoneCustomisationFutureFunc(
            customization_add_func=mock_add_func,
            order=MockOrder(),
            timeout=0.001,
            job_identificator=JobIdentifier(total_jobs=1),
        )(FakeProduct())


inactive_transitions = [
    cycle(["RUNNING", "INACTIVE"]),
    cycle(["QUEUED", "INACTIVE"]),
]


@pytest.mark.parametrize("transitions", inactive_transitions)
def test_inactive_transitions(transitions):
    with pytest.raises(EumdacError):

        def mock_add_func(product):
            return FakeCustomisation(transitions)

        fut = WaitForDoneCustomisationFutureFunc(
            customization_add_func=mock_add_func,
            order=MockOrder(),
            timeout=1000,
            job_identificator=JobIdentifier(total_jobs=1),
        )
        fut(FakeProduct())
        assert fut.state.name == "INACTIVE"


class FakeAdd:
    def __init__(self):
        self.exception_cnt = 0

    def func(self, product):
        if isinstance(product, tuple):
            exc, max_cnt = product
            if self.exception_cnt >= max_cnt:
                return FakeCustomisation([])
            self.exception_cnt += 1
            raise CustomisationError("Could not add customisation")
        return FakeCustomisation([])


eumdac_error_descriptions = [
    {
        "status": 400,
        "title": "invalid value for parameter",
        "description": "PNG (RGB) driver doesn't support 11 bands.  Must be 1 (grey) or 3 (rgb) bands",
    },
    {
        "status": 400,
        "title": "invalid value for parameter",
        "description": "The roi value can not be a list",
    },
]


@pytest.mark.parametrize("description", eumdac_error_descriptions)
def test_concurrent_download_fail_400(description):
    products = [
        FakeProduct(),
    ]
    order = MockOrder()
    output_dir = "out"
    token = None

    def mock_add_func(product):
        raise EumdacError("Test", description)

    with pytest.raises(FatalEumdacError):
        WaitForDoneCustomisationFutureFunc(
            customization_add_func=mock_add_func,
            order=MockOrder(),
            job_identificator=JobIdentifier(total_jobs=1),
        )(FakeProduct())


class FakeFailCustomisation(FakeCustomisation):
    def __init__(self, logfile, cid=None):
        self.states_to_return = iter(["QUEUED", "RUNNING", "FAILED"])
        self.logfile = logfile
        self._id = cid or "foo"
        self.deleted = False


class FakeFailCustomisationProduct:
    def __init__(self, customisation, product):
        self.customisation = customisation
        self.product = product


eumdac_error_logs = [
    """
    2022-07-14 07:37:52 - PROCESSING.chain_runner[270] - INFO - Start process "fb6bdb44"
    2022-07-14 07:37:52 - PROCESSING.chain_runner[231] - INFO - WORKER: tcp://10.0.3.141:40353
    2022-07-14 07:37:52 - PROCESSING.chain_runner[232] - INFO - PID: 26
    2022-07-14 07:37:52 - PROCESSING.chain_runner[233] - INFO - backend: epct_gis_hrit
    2022-07-14 07:37:52 - PROCESSING.chain_runner[234] - INFO - user: delahoz
    2022-07-14 07:38:04 - PROCESSING.preprocessing[56] - INFO - download from https://api.eumetsat.int/data/download/collections/EO%3AEUM%3ADAT%3AMSG%3AHRSEVIRI/products/MSG4-SEVI-MSG15-0100-NA-20220601004243.155000000Z-NA is finished
    2022-07-14 07:38:07 - PROCESSING.chain_runner[401] - INFO - customisation time: 14 - process: fb6bdb44
    2022-07-14 07:38:07 - PROCESSING.chain_runner[403] - ERROR - error at preprocessing.initialise_processing[516] while initializing the customization for product(s) [MSG4-SEVI-MSG15-0100-NA-20220601004243.155000000Z-NA] and for configuration {'product': 'HRSEVIRI_HRIT', 'format': 'png_rgb'}: invalid HRIT file name: '/var/dtws/users/delahoz/workspace/EPCT_HRSEVIRI_HRIT_fb6bdb44/decompressed_data/MSG4-SEVI-MSG15-0100-NA-20220601004243.155000000Z-NA.nat'
    """,
    """
    2022-07-14 08:14:19 - PROCESSING.api[403] - INFO - Submitted process 53a71aea - FUTURE: run_chain_async-e24f66fb1ab913b7761f9ac62b4ee281
    2022-07-14 08:14:21 - PROCESSING.chain_runner[270] - INFO - Start process "53a71aea"
    2022-07-14 08:14:21 - PROCESSING.chain_runner[231] - INFO - WORKER: tcp://10.0.3.151:33526
    2022-07-14 08:14:21 - PROCESSING.chain_runner[232] - INFO - PID: 26
    2022-07-14 08:14:21 - PROCESSING.chain_runner[233] - INFO - backend: epct_gis_eps_grib2
    2022-07-14 08:14:21 - PROCESSING.chain_runner[234] - INFO - user: delahoz
    2022-07-14 08:14:39 - PROCESSING.preprocessing[56] - INFO - download from https://api.eumetsat.int/data/download/collections/EO%3AEUM%3ADAT%3AMETOP%3AAVHRRL1/products/AVHR_xxx_1B_M03_20220709215503Z_20220709233703Z_N_O_20220709233331Z is finished
    2022-07-14 08:14:50 - PROCESSING.preprocessing[523] - INFO - Processing details - product: AVHRRL1 chain-name: None chain-details: -product: AVHRRL1 -format: GRIB-2 -compression: Internal compression (HDF5 or NetCDF4)
    2022-07-14 08:14:50 - PROCESSING.preprocessing[530] - INFO - Input products: AVHR_xxx_1B_M03_20220709215503Z_20220709233703Z_N_O_20220709233331Z.nat
    2022-07-14 08:14:51 - PROCESSING.epct_gis[991] - INFO - Expected steps: ['import', 'format']
    2022-07-14 08:14:51 - PROCESSING.epct_gis[994] - INFO - Starting step "IMPORT" 1/2 ...
    2022-07-14 08:14:54 - PROCESSING.vrt[94] - INFO - Command line and its output ...
    2022-07-14 08:14:54 - PROCESSING.epct_gis[998] - INFO - ... step "IMPORT" finished!
    2022-07-14 08:14:54 - PROCESSING.epct_gis[1006] - INFO - Starting step "FORMAT" 2/2 ...
    2022-07-14 08:14:54 - PROCESSING.vrt[94] - INFO - Command line and its output ...
    2022-07-14 08:15:05 - PROCESSING.epct_gis[802] - INFO - ... step "FORMAT" finished!
    2022-07-14 08:15:05 - PROCESSING.vrt[94] - INFO - Command line and its output ...
    ERROR 6: Source dataset must have a geotransform
    2022-07-14 08:15:05 - PROCESSING.vrt[94] - INFO - Command line and its output ...
    ERROR 6: Source dataset must have a geotransform
    2022-07-14 08:15:05 - PROCESSING.chain_runner[401] - INFO - customisation time: None - process: 53a71aea
    2022-07-14 08:15:05 - PROCESSING.chain_runner[403] - ERROR - error at epct_gis.run_chain[1014] while performing the customization for product(s) [AVHR_xxx_1B_M03_20220709215503Z_20220709233703Z_N_O_20220709233331Z.nat] and for configuration {'product': 'AVHRRL1', 'format': 'grib2', 'compression': 'internal'}: INTERNAL ERROR: see Log file
    """,
]


@pytest.mark.parametrize("log", eumdac_error_logs)
def test_wait_for_done_fail(log):
    def mock_add_func(product):
        return FakeFailCustomisation(logfile=log)

    with pytest.raises(FatalEumdacError):
        WaitForDoneCustomisationFutureFunc(
            order=MockOrder(),
            customization_add_func=mock_add_func,
            job_identificator=JobIdentifier(total_jobs=1),
        )(FakeProduct())


def test_state_query():
    collection_id = "coid"
    customisation_id = "cuid"
    product_id = "pid"

    success, customisation, product = StateQueryFutureFunc()(
        product_id,
        {"col_id": collection_id, "customisation": customisation_id},
        FakeTailor(),
        FakeStore(),
    )

    assert success
    assert customisation._id == customisation_id
    assert product._id == product_id
    assert product.collection._id == collection_id


def test_resume(tailor_order):
    ta = TailorApp(tailor_order, FakeStore(), FakeTailor())

    for pi in tailor_order.iter_product_info():
        assert pi.p_dict["server_state"] == "UNSUBMITTED"

    ta.resume()

    for pi in tailor_order.iter_product_info():
        assert pi.p_dict["server_state"] == "DONE"
        for fn, ds in pi.p_dict["download_states"].items():
            assert ds == "DOWNLOADED"


@pytest.mark.parametrize("status_code,message,description", [(200, "test", "test")])
def test_check_nonfatal_error_response(tailor_order, status_code, message, description):
    from eumdac.tailor_app import check_error_response

    extra_info = {"description": description, "status": status_code}

    err = EumdacError(message, extra_info)
    check_error_response(err)


@pytest.mark.parametrize("status_code,message,description", [(400, "test", "test")])
def test_check_fatal_error_response(tailor_order, status_code, message, description):
    from eumdac.tailor_app import check_error_response

    extra_info = {"description": description, "status": status_code}

    err = EumdacError(message, extra_info)
    with pytest.raises(FatalEumdacError):
        check_error_response(err)


@pytest.mark.parametrize(
    "status_code,message,description", [(500, "test", "You are exceeding your maximum number 123")]
)
def test_check_fatal_error_response(tailor_order, status_code, message, description):
    from eumdac.tailor_app import check_error_response

    extra_info = {"description": description, "status": status_code}

    err = EumdacError(message, extra_info)
    with pytest.raises(ExceedingNumberOfCustomisationsEumdacError):
        check_error_response(err)



import pytest

from eumdac.errors import EumdacError


def test_urls(monkeypatch, tmp_path):
    from eumdac.local_tailor import new_url_filename, all_url_filenames, resolve_url, remove_url
    from eumdac.token import URLs

    test_url_id = "test"
    test_url = URLs()

    def mock_get_url_path():
        return tmp_path

    monkeypatch.setattr("eumdac.local_tailor.get_url_path", mock_get_url_path)

    test_path = new_url_filename(test_url_id)
    assert not test_path.exists()
    with test_path.open("w") as tf:
        test_url.write(tf)
    assert test_path.exists()

    # requesting an existing id works
    new_url_filename(test_url_id)

    # resolving an existing url succeeds
    resolve_url(test_url_id)

    # remove all urls
    url_paths = all_url_filenames()
    for up in url_paths:
        remove_url(up.stem)
        assert not up.exists()

    # resolving a removed url raises
    with pytest.raises(EumdacError):
        resolve_url(test_url_id)


def test_local_tailor(monkeypatch, tmp_path):
    from eumdac.local_tailor import (
        new_local_tailor,
        get_local_tailor,
        remove_local_tailor,
        is_online,
        get_api_url,
    )
    from eumdac.token import URLs

    tailor_id = "test"
    tailor_url = "http://NON.EXISTANT:40000"

    def mock_get_url_path():
        return tmp_path

    monkeypatch.setattr("eumdac.local_tailor.get_url_path", mock_get_url_path)

    tailor_path = new_local_tailor(tailor_id, tailor_url)
    assert tailor_path.exists()

    # creating the same tailor_id works
    new_local_tailor(tailor_id, tailor_url)

    _ = get_local_tailor(tailor_id)

    assert is_online(tailor_path) == False
    assert get_api_url(tailor_path).startswith(tailor_url)

    remove_local_tailor(tailor_id)
    # retrieving removed tailor_id fails
    with pytest.raises(EumdacError):
        get_local_tailor(tailor_id)



import pytest

from eumdac.download_app import (
    _divide_into_chunks,
    _check_chunks,
    _reassemble_from_chunks,
)
from pathlib import Path


def test_divide_into_chunks():
    output = _divide_into_chunks("basedir", 2948, 1000)
    expected = {
        Path("basedir") / "chunk.0": (0, 1000),
        Path("basedir") / "chunk.1": (1000, 2000),
        Path("basedir") / "chunk.2": (2000, 2948),
    }
    assert output == expected


def test_divide_into_chunks_bigger_chunk():
    output = _divide_into_chunks("basedir", 2948, 3000)
    expected = {
        Path("basedir") / "chunk.0": (0, 2948),
    }
    assert output == expected


def test_divide_into_chunks_exact_division():
    output = _divide_into_chunks("basedir", 2000, 1000)
    expected = {
        Path("basedir") / "chunk.0": (0, 1000),
        Path("basedir") / "chunk.1": (1000, 2000),
    }
    assert output == expected


def _create_binary_file(file_name, num_bytes, bvalue=b"\0"):
    with open(file_name, "wb") as file:
        file.write(bvalue * num_bytes)


def test_check_chunks(tmp_path):
    (tmp_path / Path("testprodname")).mkdir()
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.0", 1000)
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.1", 1000)
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.2", 948)
    output = _check_chunks(
        {
            tmp_path / Path("testprodname") / "chunk.0": (0, 1000),
            tmp_path / Path("testprodname") / "chunk.1": (1000, 2000),
            tmp_path / Path("testprodname") / "chunk.2": (2000, 2948),
        }
    )
    assert output


def test_check_chunks_invalid_size(tmp_path):
    (tmp_path / Path("testprodname")).mkdir()
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.0", 999)
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.1", 1000)
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.2", 948)
    output = _check_chunks(
        {
            tmp_path / Path("testprodname") / "chunk.0": (0, 1000),
            tmp_path / Path("testprodname") / "chunk.1": (1000, 2000),
            tmp_path / Path("testprodname") / "chunk.2": (2000, 2948),
        }
    )
    assert output is False


def test_check_chunks_chunk_missing(tmp_path):
    (tmp_path / Path("testprodname")).mkdir()
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.0", 999)
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.2", 948)
    output = _check_chunks(
        {
            tmp_path / Path("testprodname") / "chunk.0": (0, 1000),
            tmp_path / Path("testprodname") / "chunk.1": (1000, 2000),
            tmp_path / Path("testprodname") / "chunk.2": (2000, 2948),
        }
    )
    assert output is False


def test_reassemble_from_chunks(tmp_path):
    (tmp_path / Path("testprodname")).mkdir()
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.0", 1000, b"\0")
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.1", 1000, b"\1")
    _create_binary_file(tmp_path / Path("testprodname") / "chunk.2", 948, b"\2")
    output = _reassemble_from_chunks(
        {
            tmp_path / Path("testprodname") / "chunk.0": (0, 1000),
            tmp_path / Path("testprodname") / "chunk.1": (1000, 2000),
            tmp_path / Path("testprodname") / "chunk.2": (2000, 2948),
        },
        tmp_path / "output",
    )
    assert output.stat().st_size == 2948
    with output.open("rb") as outf:
        bindata = outf.read()
        print(type(bindata))
        assert bindata[0] == 0
        assert bindata[1000] == 1
        assert bindata[2000] == 2



"""Module containing common data to be reused accross modules"""

from eumdac.__version__ import __title__, __documentation__, __version__

headers = {
    "referer": "EUMDAC.LIB",
    "User-Agent": str(__title__ + "/" + __version__),
}
